#ifndef _ADC_H_
#define _ADC_H_

#include <dev/adcdrv.h>

/* Max. ADC Clock Frequency (Hz) */
#define ADC_CLOCK_MAX   20000000

/* Max. normal ADC startup time (us) */
#define ADC_STARTUP_NORMAL_MAX     40
/* Max. fast ADC startup time (us) */
#define ADC_STARTUP_FAST_MAX       12



#ifdef __cplusplus
 extern "C" {
#endif

/*------------------------------------------------------------------------------
 *         Macros function of register access
 *------------------------------------------------------------------------------*/

#define ADC_GetModeReg( pAdc )                ((pAdc)->ADC_MR)

#define ADC_StartConversion( pAdc )           ((pAdc)->ADC_CR = ADC_CR_START)

#define ADC_SetCalibMode(pAdc)                  ((pAdc)->ADC_CR |= ADC_CR_AUTOCAL)

#define ADC_EnableChannel( pAdc, dwChannel )    {\
            (pAdc)->ADC_CHER = (1 << (dwChannel));\
        }

#define ADC_DisableChannel(pAdc, dwChannel)  {\
            (pAdc)->ADC_CHDR = (1 << (dwChannel));\
        }

#define ADC_EnableIt(pAdc, dwMode)            {\
            (pAdc)->ADC_IER = (dwMode);\
        }

#define ADC_DisableIt(pAdc, dwMode)           {\
            (pAdc)->ADC_IDR = (dwMode);\
        }

#define ADC_SetChannelGain(pAdc,dwMode)       {\
            (pAdc)->ADC_CGR = dwMode;\
        }

#define ADC_SetChannelOffset(pAdc,dwMode)     {\
            (pAdc)->ADC_COR = dwMode;\
        }

#define ADC_EnableDataReadyIt(pAdc)         ((pAdc)->ADC_IER = ADC_IER_DRDY)

#define ADC_GetStatus(pAdc)                 ((pAdc)->ADC_ISR)

#define ADC_GetCompareMode(pAdc)            (((pAdc)->ADC_EMR)& (ADC_EMR_CMPMODE_Msk))

#define ADC_GetChannelStatus(pAdc)          ((pAdc)->ADC_CHSR)

#define ADC_GetInterruptMaskStatus(pAdc)    ((pAdc)->ADC_IMR)

#define ADC_GetLastConvertedData(pAdc)      ((pAdc)->ADC_LCDR)


#ifdef __cplusplus
}
#endif

extern void ADC_Initialize( AT91PS_TSADC pAdc, UINT32 dwId );
extern UINT32 ADC_SetClock( AT91PS_TSADC pAdc, UINT32 dwPres, UINT32 dwMck );
extern void ADC_SetTiming( AT91PS_TSADC pAdc, UINT32 dwStartup, UINT32 dwTracking, UINT32 dwSettling );
extern void ADC_SetTrigger( AT91PS_TSADC pAdc, UINT32 dwTrgSel );
extern void ADC_SetTriggerMode(AT91PS_TSADC pAdc, UINT32 dwMode);
extern void ADC_SetLowResolution( AT91PS_TSADC pAdc, UINT32 bEnDis );
extern void ADC_SetSleepMode( AT91PS_TSADC pAdc, BOOL bEnDis );
extern void ADC_SetFastWakeup( AT91PS_TSADC pAdc, BOOL bEnDis );
extern void ADC_SetSequenceMode( AT91PS_TSADC pAdc, BOOL bEnDis );
extern void ADC_SetSequence( AT91PS_TSADC pAdc, UINT32 dwSEQ1, UINT32 dwSEQ2 );
extern void ADC_SetSequenceByList( AT91PS_TSADC pAdc, UINT8 ucChList[], UINT8 ucNumCh );
extern void ADC_SetAnalogChange( AT91PS_TSADC pAdc, BOOL bEnDis );
extern void ADC_SetTagEnable( AT91PS_TSADC pAdc, BOOL bEnDis );
extern void ADC_SetCompareChannel( AT91PS_TSADC pAdc, UINT32 dwChannel ) ;
extern void ADC_SetCompareMode( AT91PS_TSADC pAdc, UINT32 dwMode ) ;
extern void ADC_SetComparisonWindow( AT91PS_TSADC pAdc, UINT32 dwHi_Lo ) ;
extern BOOL ADC_CheckConfiguration( AT91PS_TSADC pAdc, UINT32 dwMcK ) ;
extern UINT32 ADC_GetConvertedData( AT91PS_TSADC pAdc, UINT32 dwChannel ) ;
extern void ADC_SetTsAverage(AT91PS_TSADC pADC, UINT32 dwAvg2Conv);
extern UINT32 ADC_GetTsXPosition(AT91PS_TSADC pADC);
extern UINT32 ADC_GetTsYPosition(AT91PS_TSADC pADC);
extern UINT32 ADC_GetTsPressure(AT91PS_TSADC pADC);
extern void ADC_SetTsDebounce(AT91PS_TSADC pADC, UINT32 dwTime);
extern void ADC_SetTsPenDetect(AT91PS_TSADC pADC, BOOL bEnDis);
extern void ADC_SetStartupTime( AT91PS_TSADC pAdc, UINT32 dwUs );
extern void ADC_SetTrackingTime( AT91PS_TSADC pAdc, UINT32 dwNs );
extern void ADC_SetTriggerPeriod(AT91PS_TSADC pAdc, UINT32 dwPeriod);
extern void ADC_SetTsMode(AT91PS_TSADC pADC, UINT32 dwMode);
extern void ADC_TsCalibration( AT91PS_TSADC pAdc );


#endif // _ADC_H_
