/*	$ DOOROOS : adc.c, v 0.1 2014/07/18 $	*/

/*
 * Copyright (C) 2014	Jung Byoung-Oh ibosn systems
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are not permitted completely.
 *
 */

/*
 * Module Name: adc.c
 *
 * Purpose: adc interface for dooroos
 */

#include <types.h>
#include <thread.h>
#include <memory.h>
#include <message.h>
#include <interrupt.h>
#include <device.h>
#include <error.h>

#include <window.h>

#include <mach/hal_interrupt.h>

#include <mach/SAMA5D3X.h>
#include <mach/SAMA5D3Xbd.h>

#include <mach/gpio.h>

#include <hal_panic.h>

#include <dev/drvlib.h>
#include <dev/adcdrv.h>
//#include <mach/pmc.h>
//#include <mach/tc.h>

#include <cache.h>

#include "adc.h"

#define DBG_LVL_ADC		1

#define MAX_ADC_DEV         1

#define DMA_ADDR_OFFSET         0x20000000

#define DMA_BUFFER_SIZE         (MEMPOOL_SIZE*2)    // 8K


//#define TIMER_ENABLE
typedef struct _ADC_SERVER_ 
{
    AT91PS_USART        pUSARTReg;
    UINT32              nDeviceID;
    UINT32              nInterrupt;
    
    UINT32              IntMask;
    UINT32              IntRemModemStatus;
    UINT32              BufferedCSR;

    // dma
    UINT32				nDmaChannel;
    //UINT32				nDmaChannelTx;
    AT91PS_DMAC         pDmac;
    UINT8               channel;
    
    UINT16 wValue[NUM_CHANNELS];
    UINT16 wDone;
//    UINT8  ucChNum[NUM_CHANNELS];

    UINT32              dwAdcClock;
    UINT32              EnableChannels;
#if 1
	// Buffers
#define RX_BUFFERS  2
    //PVOID               vDMACommonTxBuffer;							//!< Virtual Tx Buffer
    PVOID               vDMABuffer[RX_BUFFERS];				//!< Virtual Rx Buffers
    //PVOID               pDMACommonTxBuffer;				            //!< Physical Tx Buffer
    PVOID               pDMABuffer[RX_BUFFERS];	            //!< Physical Rx Buffers
    UINT32              DMABufferSize;

    PVOID               pCurrentDmaBuffer;
    UINT32              iCurrentRxBuffer;

	//UINT32				nDataLenTx;
	HDMA_TRANSFER_DESC  TranDesc[2];

	//HDMA_TRANSFER_DESC  TxTranDesc;
#endif

	PBYTE				pTailBuf;
	UINT32				nTailBuf;


#ifdef TIMER_ENABLE

#define ADC_TIMER_DELAY          1000
#define ADC_TIMER_BITMAP         0x100
    HANDLE       hAdcDrvTimer;

#endif

    HANDLE  hThreadWait;
    VOID *  pBufferWait;

}ADC_SERVER;


STATIC ADC_SERVER        AdcServerData[MAX_ADC_DEV];

//STATIC HDMA_TRANSFER_DESC  TranDesc[RX_BUFFERS];

// Configure PIO controllers to periph mode
static struct pio_desc ADpin_pio[ADC_CHANNEL_MAX];
static const struct pio_desc ADpin_All[] = {
    {"AD0",     AT91C_PIN_PD(20),   0, PIO_DEFAULT, PIO_PERIPH_A},
    {"AD1",     AT91C_PIN_PD(21),   0, PIO_DEFAULT, PIO_PERIPH_A},
    {"AD2",     AT91C_PIN_PD(22),   0, PIO_DEFAULT, PIO_PERIPH_A},
    {"AD3",     AT91C_PIN_PD(23),   0, PIO_DEFAULT, PIO_PERIPH_A},
    {"AD4",     AT91C_PIN_PD(24),   0, PIO_DEFAULT, PIO_PERIPH_A},
    {"AD5",     AT91C_PIN_PD(25),   0, PIO_DEFAULT, PIO_PERIPH_A},
    {"AD6",     AT91C_PIN_PD(26),   0, PIO_DEFAULT, PIO_PERIPH_A},
    {"AD7",     AT91C_PIN_PD(27),   0, PIO_DEFAULT, PIO_PERIPH_A},
    {"AD8",     AT91C_PIN_PD(28),   0, PIO_DEFAULT, PIO_PERIPH_A},
    {"AD9",     AT91C_PIN_PD(29),   0, PIO_DEFAULT, PIO_PERIPH_A},
    {"AD10",    AT91C_PIN_PD(30),   0, PIO_DEFAULT, PIO_PERIPH_A},
    {"AD11",    AT91C_PIN_PD(31),   0, PIO_DEFAULT, PIO_PERIPH_A},
};

// Configure PIO controllers to periph mode
static const struct pio_desc ADTRG_pio[] = {
    {"ADTRG",     AT91C_PIN_PD(19),   0, PIO_DEFAULT, PIO_PERIPH_A},
};

#if 0
// Configure PIO controllers to periph mode
static const struct pio_desc TIOA0_pio[] = {
    {"TIOA0",     AT91C_PIN_PD(5),   0, PIO_DEFAULT, PIO_PERIPH_B},
};
#endif

//--------------------------------------------------------------
// Defines
//--------------------------------------------------------------
/*
 * We use two ADC channel for this example:
 *    ADC_CHANNEL_5  (potentiometer)
 *    ADC_CHANNEL_15 (temperature sensor)
 */

/** ADC convention done mask */
#define ADC_DONE_MASK   ( (1<<NUM_CHANNELS) - 1 )

/** Size of the receive buffer and transmit buffer. */
#define BUFFER_SIZE     NUM_CHANNELS

/** Reference voltage for ADC, in mv */
#define VOLT_REF        (3300)

/** The maximal digital value */
#define MAX_DIGITAL     (4095)






// Sample filters
#define MAX_DIFFERENCE_CHANGE		100
#define SAMPLING_TIME				0x3FFF

//DWORD gIntrTouch        = 0;
//STATIC UINT32 ADCClock = 0;

#define ABS(a)              ((a)>=0 ? (a): -(a))

//===========================================================================

STATIC BOOL  
Dma_Init( ADC_SERVER *pAdcServer )
{
//    BOOL    bRet;
	UINT32  iController;
	UINT32  dwCfg;
	
    DEBUGMSG( DBG_LVL_ADC,"%s(%d): \n",__FUNCTION__, __LINE__);

    UINT32 DmaBuffer;
    DmaBuffer = (UINT32)MemPageAlloc( ((DMA_BUFFER_SIZE*2)/MEMPOOL_SIZE) ); //1page = 4096*2
    // ADC Buffers
    pAdcServer->vDMABuffer[0] = (PVOID)(DmaBuffer);
    pAdcServer->pDMABuffer[0] = (PVOID)(DmaBuffer + DMA_ADDR_OFFSET);
    pAdcServer->vDMABuffer[1] = (PVOID)(DmaBuffer + DMA_BUFFER_SIZE);
    pAdcServer->pDMABuffer[1] = (PVOID)(DmaBuffer + DMA_BUFFER_SIZE + DMA_ADDR_OFFSET);
    pAdcServer->DMABufferSize = DMA_BUFFER_SIZE;

    pAdcServer->nDmaChannel = DMAD_AllocateChannel(SAMA5D3_ID_ADC, DMAD_TRANSFER_MEMORY, TRUE);
DBGLINE( DBG_LVL_ADC,": pAdcServer->nDmaChannel = %x\n", pAdcServer->nDmaChannel);
    if (pAdcServer->nDmaChannel == DMAD_ALLOC_FAILED)
    {
        DEBUGMSG(DBG_LVL_ADC, "%s (%d): DMA channel allocat error\n",__FUNCTION__,__LINE__);
        return FALSE;
    }

    pAdcServer->pDmac   = AT91C_BASE_DMAC1;
    pAdcServer->channel = pAdcServer->nDmaChannel & 0x7;
    
    {
        SYSINTR_REQUEST     SysIntrRequest;
        UINT32 nSysIntr = pAdcServer->nInterrupt;
    
        // request sysintr    
        SysIntrRequest.nHalIrq = ( pAdcServer->nDmaChannel & 0x07) + (pAdcServer->nDmaChannel >> 3)*8 + IRQ_VIR_DMA_START;
        SysIntrRequest.BitMap = 1 << 1; //1<<0;
        SysIntrRequest.Flag = SYSINTR_REQ_FLAG_MULTIPLE;

        if( !KernelIoControl( IOCTRL_HAL_SYSINTR_REQUEST, &SysIntrRequest, sizeof( SYSINTR_REQUEST ), &nSysIntr, sizeof(UINT32) ) )
        {
            DEBUGMSG( 1,"%s (%d): nSysIntr = %d\n", __FUNCTION__, __LINE__, nSysIntr );
        }
    }

    /* Configure DMA RX channel */
    iController = (pAdcServer->nDmaChannel >> 3);
    dwCfg = 0
           | DMAC_CFG_SRC_PER(DMAIF_Get_ChannelNumber( iController, SAMA5D3_ID_ADC, DMAD_TRANSFER_RX ))
           | DMAC_CFG_SRC_H2SEL
           | DMAC_CFG_SOD_DISABLE
#if 0
           | DMAC_CFG_FIFOCFG_HALF_CFG;
#else
           | DMAC_CFG_FIFOCFG_ALAP_CFG;
#endif
DEBUGMSG( DBG_LVL_ADC,"%s (%d): dwCfg = %x\n", __FUNCTION__, __LINE__, dwCfg );
    DMAD_PrepareChannel( pAdcServer->nDmaChannel, dwCfg );
    //DEBUGMSG( 1,"%s (%d): nSysIntr = %d\n", __FUNCTION__, __LINE__, nSysIntr );
DEBUGMSG( DBG_LVL_ADC,"%s (%d): DmaControler_Adc = %X, DmaChannl_Adc = %X\n", __FUNCTION__, __LINE__, (pAdcServer->nDmaChannel >> 3), (pAdcServer->nDmaChannel & 0x07) );

    return(TRUE);
}




//------------------------------------------------------------------------------------------------------------
//! \brief stops PDC (DMA) transfer
//-------------------------------------------------------------------------------------------------------------
STATIC BOOL  
DMA_Start( ADC_SERVER *pAdcServer )
{
    AT91PS_TSADC pADC = (AT91PS_TSADC)AT91C_BASE_TSADC;
    HDMA_TRANSFER_DESC  *pTranDesc;

    pTranDesc = &pAdcServer->TranDesc[0];
    pTranDesc->SrcAddr = (UINT32)&pADC->ADC_LCDR;
    pTranDesc->DstAddr = (UINT32)pAdcServer->pDMABuffer[0];
    pTranDesc->CtrlA   = DMAC_CTRLA_BTSIZE(pAdcServer->DMABufferSize/2) |
                            DMAC_CTRLA_SRC_WIDTH_HALF_WORD |
                            DMAC_CTRLA_DST_WIDTH_HALF_WORD;
    pTranDesc->CtrlB   = /*DMAC_CTRLB_SRC_DSCR | DMAC_CTRLB_DST_DSCR |*/
                            DMAC_CTRLB_SIF_AHB_IF2 |
                            DMAC_CTRLB_DIF_AHB_IF0 |
                            DMAC_CTRLB_FC_PER2MEM_DMA_FC |
                            DMAC_CTRLB_SRC_INCR_FIXED |
                            DMAC_CTRLB_DST_INCR_INCREMENTING;// | DMAC_CTRLB_AUTO;
    pTranDesc->DscAddr = (UINT32)&pAdcServer->TranDesc[1] + DMA_ADDR_OFFSET;
    CacheRangeFlush(pAdcServer->pDMABuffer[0] - DMA_ADDR_OFFSET, pAdcServer->DMABufferSize, CACHE_SYNC_DISCARD);

    pTranDesc = &pAdcServer->TranDesc[1];
    pTranDesc->SrcAddr = (UINT32)&pADC->ADC_LCDR;
    pTranDesc->DstAddr = (UINT32)pAdcServer->pDMABuffer[1];
    pTranDesc->CtrlA   = DMAC_CTRLA_BTSIZE(pAdcServer->DMABufferSize/2) |
                            DMAC_CTRLA_SRC_WIDTH_HALF_WORD |
                            DMAC_CTRLA_DST_WIDTH_HALF_WORD;
    pTranDesc->CtrlB   = /*DMAC_CTRLB_SRC_DSCR | DMAC_CTRLB_DST_DSCR |*/
                            DMAC_CTRLB_SIF_AHB_IF2 |
                            DMAC_CTRLB_DIF_AHB_IF0 |
                            DMAC_CTRLB_FC_PER2MEM_DMA_FC |
                            DMAC_CTRLB_SRC_INCR_FIXED |
                            DMAC_CTRLB_DST_INCR_INCREMENTING;// | DMAC_CTRLB_AUTO;
    pTranDesc->DscAddr = (UINT32)&pAdcServer->TranDesc[0] + DMA_ADDR_OFFSET;
    CacheRangeFlush(pAdcServer->pDMABuffer[1] - DMA_ADDR_OFFSET, pAdcServer->DMABufferSize, CACHE_SYNC_DISCARD);

    CacheRangeFlush(pAdcServer->TranDesc, sizeof(HDMA_TRANSFER_DESC)*2, CACHE_SYNC_WRITEBACK);
    DMAD_PrepareMultiTransfer(pAdcServer->nDmaChannel, (HDMA_TRANSFER_DESC *)((UINT32)pAdcServer->TranDesc + DMA_ADDR_OFFSET));
    //DMAD_PrepareSingleTransfer( nDmaChannelRx, &TranDesc);
    DMAD_StartTransfer( pAdcServer->nDmaChannel );

#if 0

AT91PS_SSC pSSC = (AT91PS_SSC)AT91C_BASE_SSC0;

    switch(wapidir)
    {
    case WAPI_IN:
        if(m_InputDMARunning)
            break;

        //----- 1. Initialize our buffer counters -----
        m_InputDMARunning=TRUE;

        m_InputDMAStatus = (DMA_DONEA | DMA_DONEB) & ~DMA_BIU;

        CODEC_PowerControl();                       // Turn On Channel
        CODEC_MuteControl(DMA_CH_IN, FALSE);        // Unmute Input Channel

		TranDesc[0].SrcAddr = (UINT32)&pSSC->SSC_RHR;
		TranDesc[0].DstAddr = (UINT32)DMA_RXBuffer;
		TranDesc[0].CtrlA   = DMAC_CTRLA_BTSIZE(DMA_RXBuf_Size/2) |
							 DMAC_CTRLA_SRC_WIDTH_HALF_WORD |
							 DMAC_CTRLA_DST_WIDTH_HALF_WORD;
		TranDesc[0].CtrlB   = DMAC_CTRLB_SRC_DSCR | DMAC_CTRLB_DST_DSCR |
							 DMAC_CTRLB_FC_PER2MEM_DMA_FC |
							 DMAC_CTRLB_SRC_INCR_FIXED |
#if 0
                             DMAC_CTRLB_SIF_AHB_IF2 |
                             DMAC_CTRLB_DIF_AHB_IF0 |
#endif
							 DMAC_CTRLB_DST_INCR_INCREMENTING;
		TranDesc[0].DscAddr = (UINT32)&TranDesc[1] + DMA_ADDR_OFFSET;
		CacheRangeFlush(DMA_RXBuffer - DMA_ADDR_OFFSET, DMA_RXBuf_Size, CACHE_SYNC_DISCARD);
		if(DMA_RXBuffer2)
		{
			TranDesc[1].SrcAddr = (UINT32)&pSSC->SSC_RHR;
			TranDesc[1].DstAddr = (UINT32)DMA_RXBuffer2;
			TranDesc[1].CtrlA   = DMAC_CTRLA_BTSIZE(DMA_RXBuf2_Size/2) |
								 DMAC_CTRLA_SRC_WIDTH_HALF_WORD |
								 DMAC_CTRLA_DST_WIDTH_HALF_WORD;
			TranDesc[1].CtrlB   = DMAC_CTRLB_SRC_DSCR | DMAC_CTRLB_DST_DSCR |
								 DMAC_CTRLB_FC_PER2MEM_DMA_FC |
								 DMAC_CTRLB_SRC_INCR_FIXED |
#if 0
                                 DMAC_CTRLB_SIF_AHB_IF2 |
                                 DMAC_CTRLB_DIF_AHB_IF0 |
#endif
								 DMAC_CTRLB_DST_INCR_INCREMENTING;
			TranDesc[1].DscAddr = (UINT32)&TranDesc[0] + DMA_ADDR_OFFSET;
			CacheRangeFlush(DMA_RXBuffer2 - DMA_ADDR_OFFSET, DMA_RXBuf2_Size, CACHE_SYNC_DISCARD);
		}
		CacheRangeFlush(TranDesc, sizeof(HDMA_TRANSFER_DESC)*2, CACHE_SYNC_WRITEBACK);
		DMAD_PrepareMultiTransfer(nDmaChannelRx, (HDMA_TRANSFER_DESC *)((UINT32)TranDesc + DMA_ADDR_OFFSET));
		//DMAD_PrepareSingleTransfer( nDmaChannelRx, &TranDesc);
		DMAD_StartTransfer( nDmaChannelRx);
		pSSC->SSC_CR = SSC_CR_RXEN;
		break;

    case WAPI_OUT:    
   
        //----- 1. Initialize our buffer counters -----

		{
			if(m_OutputDMARunning)
				break;

			m_OutputDMARunning = TRUE;

#if 0	//over 3 multi decripter 
			UINT32 i =0;
			for(i=0;i<MAX_DESC_NUM;i++)
			{
				if((i%2) == 0)
				{
					TxTranDesc[i].SrcAddr = (UINT32)DMA_TXBuffer;
					TxTranDesc[i].DstAddr = (UINT32)&pSSC->SSC_THR;
					TxTranDesc[i].CtrlA   = DMAC_CTRLA_BTSIZE(DMA_TXBuf_Size/2) |
										 DMAC_CTRLA_SRC_WIDTH_HALF_WORD |
										 DMAC_CTRLA_DST_WIDTH_HALF_WORD;
					TxTranDesc[i].CtrlB   = 0|//DMAC_CTRLB_SRC_DSCR | DMAC_CTRLB_DST_DSCR |
										 DMAC_CTRLB_FC_MEM2PER_DMA_FC |
										 DMAC_CTRLB_SRC_INCR_INCREMENTING |
										 DMAC_CTRLB_DST_INCR_FIXED;
					TxTranDesc[i].DscAddr =(UINT32)&TxTranDesc[i+1] + DMA_ADDR_OFFSET;

				}
				else if((i%2) == 1)
				{

					TxTranDesc[i].SrcAddr = (UINT32)DMA_TXBuffer2;
					TxTranDesc[i].DstAddr = (UINT32)&pSSC->SSC_THR;
					TxTranDesc[i].CtrlA   = DMAC_CTRLA_BTSIZE(DMA_TXBuf2_Size/2) |
									 DMAC_CTRLA_SRC_WIDTH_HALF_WORD |
									 DMAC_CTRLA_DST_WIDTH_HALF_WORD;
					TxTranDesc[i].CtrlB   = 0|//DMAC_CTRLB_SRC_DSCR | DMAC_CTRLB_DST_DSCR |
									 DMAC_CTRLB_FC_MEM2PER_DMA_FC |
									 DMAC_CTRLB_SRC_INCR_INCREMENTING |
									 DMAC_CTRLB_DST_INCR_FIXED;
					TxTranDesc[i].DscAddr = (UINT32)&TxTranDesc[i+1] + DMA_ADDR_OFFSET;
				}
				
				if(i  == (MAX_DESC_NUM -1))
				   TxTranDesc[i].DscAddr = (UINT32)&TxTranDesc[0] + DMA_ADDR_OFFSET;

				CacheRangeFlush(&TxTranDesc[i], sizeof(HDMA_TRANSFER_DESC), CACHE_SYNC_WRITEBACK);
				CacheRangeFlush(DMA_TXBuffer - DMA_ADDR_OFFSET, DMA_TXBuf_Size, CACHE_SYNC_WRITEBACK);
				CacheRangeFlush(DMA_TXBuffer2 - DMA_ADDR_OFFSET, DMA_TXBuf2_Size, CACHE_SYNC_WRITEBACK);
				
			}
#else//2 multi decripter 
			TxTranDesc[0].SrcAddr = (UINT32)DMA_TXBuffer;
			TxTranDesc[0].DstAddr = (UINT32)&pSSC->SSC_THR;
			TxTranDesc[0].CtrlA   = DMAC_CTRLA_BTSIZE(DMA_TXBuf_Size/2) |
								 DMAC_CTRLA_SRC_WIDTH_HALF_WORD |
								 DMAC_CTRLA_DST_WIDTH_HALF_WORD;
			TxTranDesc[0].CtrlB   = 0|//DMAC_CTRLB_SRC_DSCR | DMAC_CTRLB_DST_DSCR |
								 DMAC_CTRLB_FC_MEM2PER_DMA_FC |
								 DMAC_CTRLB_SRC_INCR_INCREMENTING |
#if 1
                                 DMAC_CTRLB_SIF_AHB_IF0 |
                                 DMAC_CTRLB_DIF_AHB_IF2 |
#endif
								 DMAC_CTRLB_DST_INCR_FIXED |
								 DMAC_CTRLB_AUTO
								 ;
			TxTranDesc[0].DscAddr =(UINT32)&TxTranDesc[1] + DMA_ADDR_OFFSET;


			//CacheRangeFlush(DMA_TXBuffer - DMA_ADDR_OFFSET, DMA_TXBuf_Size, CACHE_SYNC_WRITEBACK);
			CacheRangeFlush(&TxTranDesc[0], sizeof(HDMA_TRANSFER_DESC), CACHE_SYNC_WRITEBACK);

			TxTranDesc[1].SrcAddr = (UINT32)DMA_TXBuffer2;
			TxTranDesc[1].DstAddr = (UINT32)&pSSC->SSC_THR;
			TxTranDesc[1].CtrlA   = DMAC_CTRLA_BTSIZE(DMA_TXBuf2_Size/2) |
							 DMAC_CTRLA_SRC_WIDTH_HALF_WORD |
							 DMAC_CTRLA_DST_WIDTH_HALF_WORD;
			TxTranDesc[1].CtrlB   = 0|//DMAC_CTRLB_SRC_DSCR | DMAC_CTRLB_DST_DSCR |
							 DMAC_CTRLB_FC_MEM2PER_DMA_FC |
							 DMAC_CTRLB_SRC_INCR_INCREMENTING |
#if 1
                             DMAC_CTRLB_SIF_AHB_IF0 |
                             DMAC_CTRLB_DIF_AHB_IF2 |
#endif
							 DMAC_CTRLB_DST_INCR_FIXED |
							 DMAC_CTRLB_AUTO
							 ;
			TxTranDesc[1].DscAddr = (UINT32)&TxTranDesc[0] + DMA_ADDR_OFFSET;
			//CacheRangeFlush(DMA_TXBuffer2 - DMA_ADDR_OFFSET, DMA_TXBuf2_Size, CACHE_SYNC_WRITEBACK);
			CacheRangeFlush(&TxTranDesc[1], sizeof(HDMA_TRANSFER_DESC), CACHE_SYNC_WRITEBACK);

//kprintf("TxTranDesc[0].SrcAddr = %x\n", TxTranDesc[0].SrcAddr);
//kprintf("TxTranDesc[0].DstAddr = %x\n", TxTranDesc[0].DstAddr);
//kprintf("TxTranDesc[0].CtrlA = %x\n",   TxTranDesc[0].CtrlA);
//kprintf("TxTranDesc[0].CtrlB = %x\n",   TxTranDesc[0].CtrlB);
//kprintf("TxTranDesc[0].DscAddr = %x\n", TxTranDesc[0].DscAddr);
//
//kprintf("TxTranDesc[1].SrcAddr = %x\n", TxTranDesc[1].SrcAddr);
//kprintf("TxTranDesc[1].DstAddr = %x\n", TxTranDesc[1].DstAddr);
//kprintf("TxTranDesc[1].CtrlA = %x\n",   TxTranDesc[1].CtrlA);
//kprintf("TxTranDesc[1].CtrlB = %x\n",   TxTranDesc[1].CtrlB);
//kprintf("TxTranDesc[1].DscAddr = %x\n", TxTranDesc[1].DscAddr);

#endif
    
            m_OutputDMAStatus = (DMA_DONEA | DMA_DONEB) & ~DMA_BIU;
            //OutputTransferred = TransferOutputBuffer(m_OutputDMAStatus);
			CODEC_PowerControl();   
            // Turn Output Channel
			CODEC_MuteControl(DMA_CH_OUT, FALSE);    // nmute Output Channel

			DMAD_PrepareMultiTransfer(nDmaChannelTx, (HDMA_TRANSFER_DESC *)((UINT32)&TxTranDesc[0] + DMA_ADDR_OFFSET));

			//DMAD_PrepareSingleTransfer( nDmaChannelTx, &TxTranDesc[0]);
			DMAD_StartTransfer( nDmaChannelTx);
			pSSC->SSC_CR = SSC_CR_TXEN;
//for(;;)
//{
//    short i;
////kprintf("pSSC->SSC_TCMR    = %x\n", pSSC->SSC_TCMR);
////kprintf("pSSC->SSC_TFMR    = %x\n", pSSC->SSC_TFMR);
////kprintf("pSSC->SSC_TSHR    = %x\n", pSSC->SSC_TSHR);
////kprintf("pSSC->SSC_SR      = %x\n", pSSC->SSC_SR);
//
//pSSC->SSC_THR = i++;
//DELAY(1);
////kprintf("pSSC->SSC_SR      = %x\n", pSSC->SSC_SR);
//
//}
		}
        break;
        
    default:
//START_ERROR:
        return FALSE;
    }
#endif
    return(TRUE);
}



#if 0
/**
 * \brief Configure to trigger ADC by TIOA output of timer.
 */
static void _ConfigureTimerTrigger( void )
{
    UINT32 dwDiv;
    UINT32 dwTcClks;
    UINT32 dwMaxTcClks;

    /* PIO configuration */
    //PIO_Configure( &gPinTIOA0, PIO_LISTSIZE( gPinTIOA0 ) ) ;
    pio_setup(TIOA0_pio, sizeof(TIOA0_pio)/sizeof(struct pio_desc));

    /* Enable peripheral clock. */
    dwMaxTcClks = PMC_SetPeriMaxClock(SAMA5D3_ID_TC0, BOARD_MCK);
    PMC_EnablePeripheral( SAMA5D3_ID_TC0 ) ;

    /* Configure TC for a 1Hz frequency and trigger on RC compare. */
    TC_FindMckDivisor( 1, dwMaxTcClks, &dwDiv, &dwTcClks, dwMaxTcClks ) ;
    TC_Configure( AT91C_BASE_TCB, 0, dwTcClks | TC_CMR_CPCTRG | TC_CMR_WAVE | TC_CMR_ACPA_CLEAR | TC_CMR_ACPC_SET) ;
    AT91C_BASE_TCB->TCB_CHANNEL[0].TC_RA = ( dwMaxTcClks / dwDiv ) / 2 ;
    AT91C_BASE_TCB->TCB_CHANNEL[0].TC_RC = ( dwMaxTcClks / dwDiv ) / 1 ;

    /* Start the Timer */
    TC_Start( AT91C_BASE_TCB, 0 ) ;

    /* Set TIOA0 trigger */
    ADC_SetTrigger( AT91C_BASE_TSADC, ADC_MR_TRGSEL_ADC_TRIG1 ) ;
    ADC_SetTriggerMode (AT91C_BASE_TSADC, ADC_TRGR_TRGMOD_EXT_TRIG_ANY); /* External Trigger Any Edge */

    return;
}
#endif

STATIC INT32 Adc_Control_Mode( ADC_SERVER *pAdcServer, ADC_MODE_CONTROL *pAdcModeControl )
{
    INT32   ret = 0;

    pAdcServer->EnableChannels = pAdcModeControl->EnableChannels;

    if( pAdcServer->EnableChannels == 0 )
        return 0;

    // make pio setup
    INT32   iChannel;
    INT32   tChannel = 0;
    for( iChannel = 0; iChannel < ADC_CHANNEL_MAX; iChannel++ )
    {
        if( pAdcServer->EnableChannels & (1<<iChannel) )
        {
            ADpin_pio[tChannel] = ADpin_All[iChannel];
            tChannel++;
        }
    }

    pio_setup(ADpin_pio, tChannel);

    /* Enable/disable sequencer */
    if ( pAdcModeControl->bSequenceEn )
    {
        UINT32  Seq1 = 0;
        UINT32  Seq2 = 0;
        UINT32  iOffset = 0;

        for( iChannel = 0; iChannel < ADC_CHANNEL_MAX; iChannel++ )
        {
            if( (pAdcServer->EnableChannels & (1<<iChannel)) == 0)
                continue;

            if( iOffset > 32)
            {
                Seq2 = ADC_CHANNEL_11 << (iOffset-32); 
                iOffset += 4;
            }
            else
            {
                Seq1 = ADC_CHANNEL_1 << iOffset;
                iOffset += 4;
            }
        }

        /* Set user defined channel sequence */
        ADC_SetSequence( AT91C_BASE_TSADC, Seq1, Seq2 ) ;
        /* Enable sequencer */
        ADC_SetSequenceMode(AT91C_BASE_TSADC, 1 ) ;

        /* Enable channels */
        for( iChannel = 0; iChannel < ADC_CHANNEL_MAX; iChannel++ )
        {
            if( (pAdcServer->EnableChannels & (1<<iChannel)) == 0)
                continue;

            ADC_EnableChannel( AT91C_BASE_TSADC, iChannel ) ;
        }
        /* Update channel number */
        //pAdcServer->ucChNum[0] = ADC_CHANNEL_11 ;
        //pAdcServer->ucChNum[1] = ADC_CHANNEL_2 ;
        //pAdcServer->ucChNum[2] = ADC_CHANNEL_0 ;
    }
    else
    {
        /* Disable sequencer */
        ADC_SetSequenceMode( AT91C_BASE_TSADC, 0 ) ;

        /* Enable  channels */
        for( iChannel = 0; iChannel < ADC_CHANNEL_MAX; iChannel++ )
        {
            if( (pAdcServer->EnableChannels & (1<<iChannel)) == 0)
                continue;

            ADC_EnableChannel( AT91C_BASE_TSADC, iChannel ) ;
        }

        /* Update channel number */
        //pAdcServer->ucChNum[0] = ADC_CHANNEL_0 ;
        //pAdcServer->ucChNum[1] = ADC_CHANNEL_2 ;
        //pAdcServer->ucChNum[2] = ADC_CHANNEL_11 ;
    }

    /* Set gain and offset (only single ended mode used here) */
    ADC_SetAnalogChange( AT91C_BASE_TSADC, 0 ) ; /* Disable analog change */
    if ( pAdcModeControl->bGainEn  )
    {
        ADC_SetAnalogChange( AT91C_BASE_TSADC, 1) ;
        ADC_SetChannelGain( AT91C_BASE_TSADC, ADC_CGR_GAIN5(2) ) ; /* gain = 2 */
    }
    else
    {
        ADC_SetChannelGain( AT91C_BASE_TSADC, 0 ) ; /* gain = 1 */
    }

    if ( pAdcModeControl->bOffsetEn )
    {
        ADC_SetAnalogChange( AT91C_BASE_TSADC, 1 ) ;
        ADC_SetChannelOffset( AT91C_BASE_TSADC, ADC_COR_OFF5 ) ;
    }
    else
    {
        ADC_SetChannelOffset( AT91C_BASE_TSADC, 0 ) ;
    }
    /* Set Auto Calibration Mode*/
    if(pAdcModeControl->bAutoCalibEn)
    {
        ADC_SetCalibMode(AT91C_BASE_TSADC);
        while(1)
        {
            if((ADC_GetStatus( AT91C_BASE_TSADC ) & ADC_ISR_EOCAL) == ADC_ISR_EOCAL)    break;
        }
    }
    /* Set power save */
    if ( pAdcModeControl->bPowerSaveEn )
    {
        ADC_SetSleepMode( AT91C_BASE_TSADC, 1 ) ;
    }
    else
    {
        ADC_SetSleepMode( AT91C_BASE_TSADC, 0 ) ;
    }

    /* Transfer with/without DMA */
    if ( pAdcModeControl->bDmaEn )
    {
//        _AdcDma( dmaRxChannel, (void*)gAdcSampleData.wValue, NUM_CHANNELS );
        DMA_Start(pAdcServer);
    }
    else
    {
        //last channel;
        UINT32  LastChannel = 0;
        
        /* Enable Data ready interrupt */
        for( iChannel = 0; iChannel < ADC_CHANNEL_MAX; iChannel++ )
        {
            if( (pAdcServer->EnableChannels & (1<<iChannel)) == 0)
                continue;

            LastChannel = iChannel;
        }
        ADC_EnableIt( AT91C_BASE_TSADC, 1<<LastChannel ) ;
    }

    /* Enable ADC interrupt */
    //IRQ_ConfigureIT( ID_ADC, 0, ADC_IrqHandler );
    //IRQ_EnableIT( ID_ADC );

    /* Configure trigger mode and start convention */
    switch ( pAdcModeControl->TriggerMode )
    {
        case TRIGGER_MODE_SOFTWARE:
            ADC_SetTrigger( AT91C_BASE_TSADC, 0 ) ; /* Disable hardware trigger */
            ADC_SetTriggerMode (AT91C_BASE_TSADC, ADC_TRGR_TRGMOD_NO_TRIGGER); /* No trigger, only software trigger can start conversions */
            break ;

        case TRIGGER_MODE_ADTRG:
            //PIO_Configure( &gPinADTRG, 1 ) ;
            //PIO_Configure( &pinTrg, 1 ) ;
            pio_setup(ADTRG_pio, sizeof(ADTRG_pio)/sizeof(struct pio_desc));
            ADC_SetTrigger( AT91C_BASE_TSADC, ADC_MR_TRGSEL_ADC_TRIG0 ) ;
            ADC_SetTriggerMode (AT91C_BASE_TSADC, ADC_TRGR_TRGMOD_EXT_TRIG_ANY); /* External Trigger Any Edge */
            break ;
#if 0
        case TRIGGER_MODE_TIMER :
            //DrvTimerStart(pAdcServer->hAdcDrvTimer);
            _ConfigureTimerTrigger() ;
            break ;
#endif
        case TRIGGER_MODE_PERIOD:
        {
            ADC_SetTrigger( AT91C_BASE_TSADC, 0 ) ; /* Disable hardware trigger */
            ADC_SetTriggerPeriod(AT91C_BASE_TSADC, pAdcModeControl->Period /*period*/);
            ADC_SetTriggerMode (AT91C_BASE_TSADC, ADC_TRGR_TRGMOD_PERIOD_TRIG); /* periodic  */
            break;
        }

        default :
            break ;
    }


    return ret;
}



STATIC VOID
Adc_Init( ADC_SERVER *pAdcServer )
{
//    UINT32 dw ;
    UINT32 dwAdcClock ;
    UINT32 dwMaxAdcClock ;

    DEBUGMSG(1, "%s (%d): ========= Adc_Init\n", __FUNCTION__, __LINE__);


    pAdcServer->EnableChannels = 0;

    /* Stop */
//    DMAD_StopTransfer( &dmad, dmaRxChannel );
    ADC_DisableIt( AT91C_BASE_TSADC, 0xFFFFFFFF );

    /* Initialize ADC */
    ADC_Initialize( AT91C_BASE_TSADC, SAMA5D3_ID_ADC );
//    memset( (void*)&gAdcSampleData, 0, sizeof( gAdcSampleData ) );
//    CP15_coherent_dcache_for_dma((UINT32)&gAdcSampleData, ((UINT32)(&gAdcSampleData) + sizeof( gAdcSampleData )));
    /*
     * Formula: ADCClock = MCK / ( (PRESCAL+1) * 2 )
     * For example, MCK = 64MHZ, PRESCAL = 4, then:
     *     ADCClock = 64 / ((4+1) * 2) = 6.4MHz;
     */
    /* Set ADC clock */
    dwMaxAdcClock = PMC_SetPeriMaxClock(SAMA5D3_ID_ADC, BOARD_MCK);

    // Enable peripheral clock
    PMC_EnablePeripheral( SAMA5D3_ID_ADC ) ;

    dwAdcClock = ADC_SetClock( AT91C_BASE_TSADC, 0, dwMaxAdcClock ) ;
    kprintf( "Set ADC clock to %d Hz\n\r", (int)dwAdcClock ) ;


    pAdcServer->dwAdcClock = dwAdcClock;
    /* Formula:
     *     Startup  Time = startup value / ADCClock
     *     Transfer Time = (TRANSFER * 2 + 3) / ADCClock
     *     Tracking Time = (TRACKTIM + 1) / ADCClock
     *     Settling Time = settling value / ADCClock
     * For example, ADC clock = 6MHz (166.7 ns)
     *     Startup time = 512 / 6MHz = 85.3 us
     *     Transfer Time = (1 * 2 + 3) / 6MHz = 833.3 ns
     *     Tracking Time = (0 + 1) / 6MHz = 166.7 ns
     *     Settling Time = 3 / 6MHz = 500 ns
     */
    /* Set ADC timing */
    ADC_SetTiming( AT91C_BASE_TSADC, ADC_MR_STARTUP_SUT512, ADC_MR_TRACKTIM( 0 ), ADC_MR_SETTLING_AST17 ) ;

    /* Enable channel number tag */
    ADC_SetTagEnable( AT91C_BASE_TSADC, 1 ) ;



    /* initialize the interrupt */
{
    SYSINTR_REQUEST     SysIntrRequest;
    UINT32              nSysIntr = 0;

    // request sysintr    
    SysIntrRequest.nHalIrq = SAMA5D3_ID_ADC;
    SysIntrRequest.BitMap = 1<<0;
    SysIntrRequest.Flag = SYSINTR_REQ_FLAG_STATIC;
DEBUGMSG( DBG_LVL_ADC,"%s (%d): SysIntrRequest.nHalIrq = %d\n", __FUNCTION__, __LINE__, SysIntrRequest.nHalIrq);

    if( !KernelIoControl( IOCTRL_HAL_SYSINTR_REQUEST, &SysIntrRequest, sizeof( SYSINTR_REQUEST ), &nSysIntr, sizeof(UINT32) ) )
    {
        DEBUGMSG( DBG_LVL_ADC,"%s (%d): Error, nSysIntr = %d\n", __FUNCTION__, __LINE__, nSysIntr );
        return;
    }
	DEBUGMSG( DBG_LVL_ADC,"%s (%d): nSysIntr = %d\n", __FUNCTION__, __LINE__, nSysIntr);
	
    pAdcServer->nInterrupt = nSysIntr;
}

	return;
}

STATIC UINT32       Channel;

STATIC ADC_MODE_CONTROL DefaultAdcModeControl = 
{
    TRIGGER_MODE_SOFTWARE, //TRIGGER_MODE   TriggerMode;
    //TRIGGER_MODE_PERIOD, //TRIGGER_MODE   TriggerMode;
    FALSE,      //BOOL          bDmaEn;
    FALSE,      //BOOL          bSequenceEn;
    FALSE,      //BOOL          bGainEn;
    FALSE,      //BOOL          bOffsetEn;
    FALSE,      //BOOL          bPowerSaveEn;
    FALSE,      //BOOL          bAutoCalibEn;
    312500,
    //212000,    // UINT32       Period in ns
                //UINT32        EnableChannels
        0
//    | (1<<ADC_CHANNEL_0)
//    | (1<<ADC_CHANNEL_1)
//    | (1<<ADC_CHANNEL_2) 
//    | (1<<ADC_CHANNEL_3) 
//    | (1<<ADC_CHANNEL_4) 
//    | (1<<ADC_CHANNEL_5) 
//    | (1<<ADC_CHANNEL_6) 
//    | (1<<ADC_CHANNEL_7) 
//    | (1<<ADC_CHANNEL_8) 
//    | (1<<ADC_CHANNEL_9) 
//    | (1<<ADC_CHANNEL_10)
//      (1<<ADC_CHANNEL_11)
};


STATIC INT32
th_AdcDevice( UINT32 Arg )
{
    JB_RECVMESSAGE      Message;
    HANDLE              hPort;
    HANDLE              Handle = NULL;
    PDEVICECONFIG       pDeviceConfig;

    ADC_SERVER*         pAdcServer;

    pDeviceConfig = (PDEVICECONFIG)Arg;
    
    hPort = PortGet( );    
    PortSetType( hPort, PORTTYPE_MESSAGE| PORTTYPE_INTERRUPT| PORTTYPE_SIGNAL );

    Channel = ChannelGet( hPort );

    pAdcServer = AdcServerData + pDeviceConfig->Minor;

    pDeviceConfig->Name[3] = (UINT32)pDeviceConfig->Minor + '1';
    pDeviceConfig->Channel = Channel;
    DeviceRegister( (PDEVICECONFIG)pDeviceConfig );

    Adc_Init( pAdcServer );
    Dma_Init( pAdcServer );

    /* set default */
    Adc_Control_Mode(pAdcServer, &DefaultAdcModeControl);

    InterruptRegister(pAdcServer->nInterrupt, hPort);
    InterruptEnable( pAdcServer->nInterrupt );
DBGLINE( DBG_LVL_ADC,": nSysIntr = %d\n", pAdcServer->nInterrupt);

    
#ifdef TIMER_ENABLE
    pAdcServer->hAdcDrvTimer = DrvTimerCreate( Channel, ADC_TIMER_DELAY, ADC_TIMER_BITMAP );
    // start the timer
    if( pAdcServer->hAdcDrvTimer != NULL )
        DrvTimerStart(pAdcServer->hAdcDrvTimer); 
#endif

	while(1)
	{
		MessageReceive( hPort, &Message);

		switch(Message.Type)
		{
		case MESSAGE_TYPE_INTERRUPT:
		{
//            DBGLINE(1,": ADC MESSAGE_TYPE_INTERRUPT = %d\n",Message.Param);
//{
//static int count = 0;
//count++;
//if( !(count % 10000) ) kprintf("*");
//}
            if( Message.Param & (1<<0) )    // ADC_IRQ
            {
                UINT32 dwStatus;
                dwStatus = ADC_GetStatus( AT91C_BASE_TSADC );
//DBGLINE(1,": dwStatus = %x\n",dwStatus);
                int i;
                for( i = ADC_CHANNEL_0; i <= ADC_CHANNEL_11; i++ )
                {
                    if ((dwStatus & (1<<i) ) == (1<<i)) 
                    {
                        pAdcServer->wValue[i] = ADC_GetConvertedData(AT91C_BASE_TSADC, i);
                        pAdcServer->wDone |= (1<<i) ;
//DBGLINE(1,": pAdcServer->wValue[i] = %x\n",pAdcServer->wValue[i]);
                    } 
                }

                if( pAdcServer->hThreadWait != NULL )
                {
                    Message.hSender = pAdcServer->hThreadWait;
                    Message.Type = ADC_CHANNEL_MAX*2;
                    memcpy( (char *)pAdcServer->pBufferWait, (char *)pAdcServer->wValue, ADC_CHANNEL_MAX*2); // 16bit
                    MessageReply(&Message);
                    pAdcServer->hThreadWait = NULL; // reset
                }
            }

            if( Message.Param & (1<<1) )    // DMA_IRQ
            {
                VOID * CurrentDmaAddr = (VOID *)DMAC_GetDestinationAddr( pAdcServer->pDmac,pAdcServer->channel);
                
                if( CurrentDmaAddr < pAdcServer->pDMABuffer[1] )
                    CurrentDmaAddr = pAdcServer->pDMABuffer[1];
                else
                    CurrentDmaAddr = pAdcServer->pDMABuffer[0];
                UINT32 Ret = DMAD_Handler( pAdcServer->nDmaChannel );
                //if( Ret != DMAD_OK )
                if( Ret == DMAD_ERROR )
                {
//DBGLINE(1,": DMAD_Handler : Ret = %x\n",Ret );
                    DMAD_StopTransfer( pAdcServer->nDmaChannel );
                    DMA_Start( pAdcServer );
                }

//DBGLINE(1," : pAdcServer->hThreadWait = %d\n", pAdcServer->hThreadWait);
                if( pAdcServer->hThreadWait != NULL )
                {
                    Message.hSender = pAdcServer->hThreadWait;
                    Message.Type = pAdcServer->DMABufferSize;
                    memcpy( (char *)pAdcServer->pBufferWait, (char *)CurrentDmaAddr, pAdcServer->DMABufferSize );
                    MessageReply(&Message);
                    pAdcServer->hThreadWait = NULL; // reset
                }

                pAdcServer->pCurrentDmaBuffer = CurrentDmaAddr;
//DBGLINE(1,": DMAD_Handler CurrentDmaAddr = %x\n",*(UINT16 *)CurrentDmaAddr );
                //DMAD_Handler( &dmad );
            }

            InterruptDone(pAdcServer->nInterrupt);

            continue;
        }

		case MESSAGE_TYPE_SIGNAL:
		{

#ifdef TIMER_ENABLE		
DBGLINE(1,": Message.Param = %x\n", Message.Param);
            if( Message.Param & ADC_TIMER_BITMAP )
            {
DBGLINE(1,": ADC_StartConversion : IMR = %x\n", ADC_GetInterruptMaskStatus(AT91C_BASE_TSADC));
DBGLINE(1,": ADC_StartConversion : ISR = %x\n", ADC_GetStatus(AT91C_BASE_TSADC));
                ADC_StartConversion( AT91C_BASE_TSADC );
            }
#endif
            continue;
		}
		
		case MESSAGE_TYPE_DEVICE_READ:
        {
		    Message.Type = 0;
			break;
	    }

		case MESSAGE_TYPE_DEVICE_WRITE:
	    {
            Message.Type = 0;
            break;
	    }

		case MESSAGE_TYPE_DEVICE_CONTROL:
	    {
            UINT32      Ret = 0;
            PIOCONTROL_PARAM    pIoControlParam;
            if( Handle == NULL || Message.Handle != Handle )
            {
                ThreadSetLastError( Message.hSender, ERR_INVALID_HANDLE );
                Message.Type = FALSE;
                break;
            }

            pIoControlParam = (PIOCONTROL_PARAM)Message.Param;
		    switch( pIoControlParam->Cmd )
		    {
		    case IOCTRL_ADC_MODE:
                Adc_Control_Mode( pAdcServer, (ADC_MODE_CONTROL *)pIoControlParam->pInBuf );
		    	//PIO_Power_SetState(POWER_MAINBOARD_5V, bLedOn?FALSE:TRUE);
				//DBGLINE(1,"%s (%d) :  IOCTRL_POWER_5V = %d\n", __FUNCTION__, __LINE__,bLedOn);
                Ret = TRUE;
				break;
		    case IOCTRL_ADC_BUFFER:
                //DBGLINE(1,"%s (%d) :  IOCTRL_POWER_9V = %d\n", __FUNCTION__, __LINE__,~bLedOn);
		    	//PIO_Power_SetState(POWER_MAINBOARD_9V, bLedOn?FALSE:TRUE);
                Ret = TRUE;
				break;
		    case IOCTRL_ADC_READ_SINGLE:
                ADC_StartConversion( AT91C_BASE_TSADC );

                pAdcServer->hThreadWait = Message.hSender;
                pAdcServer->pBufferWait = pIoControlParam->pOutBuf;
                if( pAdcServer->pBufferWait == NULL )
                {
                    Message.Type = 0;
                    ThreadSetLastError( Message.hSender, ERR_INVALID_PARAMETER );
                    break;
                }
                continue;

		    case IOCTRL_ADC_READ_BUFFER:
                pAdcServer->hThreadWait = Message.hSender;
                pAdcServer->pBufferWait = pIoControlParam->pOutBuf;
//DBGLINE(1," : pAdcServer->hThreadWait = %d\n", pAdcServer->hThreadWait);
                if( pAdcServer->pBufferWait == NULL )
                {
                    Message.Type = 0;
                    ThreadSetLastError( Message.hSender, ERR_INVALID_PARAMETER );
                    break;
                }
                continue;

            default:
                ThreadSetLastError( Message.hSender, ERR_INVALID_PARAMETER );
				break;
			}

            Message.Type = Ret;
			break;
	    }

		case MESSAGE_TYPE_DEVICE_OPEN:
        {
			Handle = ChannelOpenHandleGetFrom(Channel);
			Message.Type = Handle;
            break;
        }

		case MESSAGE_TYPE_DEVICE_CLOSE:
        {
		    Handle = NULL;
		    Message.Type = TRUE;
			break;
	    }

        default:
            break;
		}

        MessageReply(&Message);

	}

	return TRUE;	
}



//==========================================================================
#define ADCDEV_STACK_SIZE       (4096)

STATIC DEVICECONFIG             AdcDeviceConfig ={"ADC", 0, -1, NULL, NULL };

STATIC THREADCONFIG             AdcDev_Thread =
                                            {
                                                (UINT32)&AdcDeviceConfig,
                                                th_AdcDevice,
                                                (PBYTE)NULL,
                                                ADCDEV_STACK_SIZE,
                                                THREAD_PRIORITY_DRIVER_HIGH+11
                                            };

//==========================================================================

#include <moduleloader.h>

PUBLIC HANDLE 
DllMain( HANDLE hDll, UINT32 Reason, PVOID Reserved )
{
    HANDLE hThread = NULL;
    switch( Reason )
    {
    case DLL_PROCESS_ATTACH:
        hThread = ThreadCreate (&AdcDev_Thread);
    	break;
    
    case DLL_PROCESS_DETACH:
    	break;
    }

    return hThread;
}

